/*******************************************************************
* DivXripper - by Hellstorm                                        *
* licensed under the GNU General Public License (GPL) v2           *
* play.cpp - player module                                         *
*******************************************************************/

#include "play_def.h"
#include "mpeg2.h"
#include <SDL.h>
#include <SDL_audio.h>
#include <SDL_timer.h>
#include <SDL_thread.h>
#include <stdlib.h>
#include <string.h>

#define NUM_SAMPLES 4096
#define NUM_SAMPLES_BUFFER 32768
#define BYTES_PER_SAMPLE 2
#define NUM_CHANNELS 2

static int end_play;
static int start_play;
static int break_play;
static int thread_exit;
static char buffer[NUM_SAMPLES_BUFFER * BYTES_PER_SAMPLE * NUM_CHANNELS];
static char buffer2[NUM_SAMPLES_BUFFER * BYTES_PER_SAMPLE * NUM_CHANNELS];
static int active_buffer = 0;
static int empty_buffer = 0;
static int buffer_status = 0;
static SDL_Thread *thread;
static int numSamples = 0;
static int numSamplesPlayed = 0;
static Uint32 startTick;
static playAudioCallback *stAudioCallback;
static unsigned char **bufferPtr;
static SDL_Surface *scr;
static int depth;
static int width, height, crop;
static int orig_width, orig_height, scrop;
static unsigned char *lastRow;
static int numFrames;
static int dropFrames;
static int astream, vstream;
static int hasAudio = 1, hasVideo = 1;
static SDL_Overlay *overlay;

int playAudioBufferFiller(void *data)
{
	int result = 0;
	playAudioCallback *stAudioCallback = (playAudioCallback *) data;
	int interval = 10;
	printf("Activation interval:%d\n", interval);

	while (!thread_exit) {
		while (empty_buffer == 0 && !thread_exit) {
			SDL_Delay(interval);
		}
		empty_buffer = 0;
		if (!thread_exit) {
			if (active_buffer == 0) {
				result =
				    mpeg2AudioRead(stAudioCallback->file,
						   buffer2,
						   NUM_SAMPLES_BUFFER,
						   stAudioCallback->
						   astream);
			} else {
				result =
				    mpeg2AudioRead(stAudioCallback->file,
						   buffer,
						   NUM_SAMPLES_BUFFER,
						   stAudioCallback->
						   astream);
			}
			numSamples += NUM_SAMPLES_BUFFER;
			if (result)
				end_play = 1;
		}
	}
	return 0;
}

void playInitAudioBuffers(mpeg3_t * file, int astream)
{
	int result;

	if (!hasAudio)
		return;

	result = mpeg2AudioRead(file, buffer, NUM_SAMPLES_BUFFER, astream);
	if (result)
		end_play = 1;
	result = mpeg2AudioRead(file, buffer2, NUM_SAMPLES_BUFFER,
				astream);
	if (result)
		end_play = 1;
	numSamples += NUM_SAMPLES_BUFFER * 2;
}

int playEventsHandler(const SDL_Event * event)
{
	if (event->type == SDL_QUIT) {
		end_play = 1;
		return 0;
	}
	return 1;
}

void playAudio(void *udata, Uint8 * stream, int len)
{
	playAudioCallback *stAudioCallback = (playAudioCallback *) udata;

	if (!thread_exit) {
		if (active_buffer == 0)
			SDL_MixAudio(stream,
				     (Uint8 *) (buffer + buffer_status),
				     len, SDL_MIX_MAXVOLUME);
		else
			SDL_MixAudio(stream,
				     (Uint8 *) (buffer2 + buffer_status),
				     len, SDL_MIX_MAXVOLUME);
		buffer_status += len;
		numSamplesPlayed += len / 4;
		if (buffer_status ==
		    NUM_SAMPLES_BUFFER * BYTES_PER_SAMPLE * NUM_CHANNELS) {
			buffer_status = 0;
			active_buffer++;
			active_buffer %= 2;
			empty_buffer = 1;
		}
		Uint32 actualTick = SDL_GetTicks();
		printf
		    ("Movie time elapsed:%fs Real time:%f Frames:%d Droped:%d\r",
		     (float) numSamplesPlayed /
		     mpeg2AudioSampleRate(stAudioCallback->file,
					  stAudioCallback->astream),
		     (actualTick - startTick) / 1000.0, numFrames,
		     dropFrames);
		fflush(stdout);
	} else {
		empty_buffer = 1;
	}
}

void playInitAudio(mpeg3_t * file, int astream)
{
	SDL_AudioSpec fmt;

	if (mpeg2NumAudioStreams(file) == 0) {
		hasAudio = 0;
		return;
	}

	stAudioCallback = new playAudioCallback;
	if (!stAudioCallback) {
		fprintf(stderr, "Not enough memory at playInit\b");
		exit(1);
	}
	stAudioCallback->file = file;
	stAudioCallback->astream = astream;

	fmt.freq = mpeg2AudioSampleRate(file, astream);
	fmt.format = AUDIO_S16;
	fmt.channels = 2;
	fmt.samples = NUM_SAMPLES;
	fmt.callback = playAudio;
	fmt.userdata = (void *) stAudioCallback;

	/* Open the audio device and start playing sound! */
	if (SDL_OpenAudio(&fmt, NULL) < 0) {
		fprintf(stderr, "Unable to open audio: %s\n",
			SDL_GetError());
		exit(1);
	}
	playInitAudioBuffers(file, astream);
	thread = SDL_CreateThread(playAudioBufferFiller, stAudioCallback);

}

int playGetBitsPerPixel()
{
	const SDL_VideoInfo *videoInfo = SDL_GetVideoInfo();
	int depth = videoInfo->vfmt->BitsPerPixel;

	if (depth < 16) {
		fprintf(stderr,
			"This player doesn't work with less than 16bpp\n");
		exit(1);
	}
	return depth;
}

void playInitVideo(mpeg3_t * file, int vstream, float ratio, int pCrop,
		   double cropRatio)
{
	int i;

	if (mpeg2NumVideoStreams(file) == 0) {
		hasVideo = 0;
		return;
	}

	orig_width = mpeg2VideoWidth(file, vstream);
	orig_height = mpeg2VideoHeight(file, vstream);

	width = orig_width;
	height = (int) (width / ratio + 0.5);

	if (cropRatio != -1) {
		crop = (int) ((height - width / cropRatio) / 2.0 + 0.5);
		scrop =
		    (int) ((double) crop / (double) height *
			   (double) orig_height + .5);
		orig_height -= scrop * 2;
		height -= crop * 2;

	} else if (pCrop != -1) {
		scrop = pCrop;
		crop =
		    (int) ((double) scrop / (double) orig_height *
			   (double) height + .5);
		orig_height -= scrop * 2;
		height -= crop * 2;
	}


	depth = playGetBitsPerPixel();
	scr = SDL_SetVideoMode(width, height, depth, SDL_HWSURFACE);
	if (scr == NULL) {
		fprintf(stderr, "Display could not be created\n");
		exit(1);
	}

	overlay =
	    SDL_CreateYUVOverlay(orig_width, orig_height, SDL_IYUV_OVERLAY,
				 scr);
}

void playReadVideoFrame(mpeg3_t * file, int vstream)
{
	if (!hasVideo)
		return;

	SDL_LockYUVOverlay(overlay);
	mpeg2VideoReadYUVFrame(file, (char *) overlay->pixels[0],
			       (char *) overlay->pixels[1],
			       (char *) overlay->pixels[2], 0, scrop,
			       orig_width, orig_height, vstream);
	SDL_UnlockYUVOverlay(overlay);
}

void playInit(mpeg3_t * file, mpeg3_t * file2, int pAstream, int pVstream,
	      float ratio, int crop, double cropRatio)
{
	if (SDL_Init(SDL_INIT_AUDIO | SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Unable to init SDL: %s\n",
			SDL_GetError());
		exit(1);
	}

	astream = pAstream;
	vstream = pVstream;

	end_play = 0;
	start_play = 0;
	break_play = 0;
	thread_exit = 0;
	numFrames = 0;
	dropFrames = 0;

	SDL_SetEventFilter(playEventsHandler);

	playInitAudio(file2, astream);
	playInitVideo(file, vstream, ratio, crop, cropRatio);

	float frameRate;
	int frameDelay, framesDrop, delay, oldDelay;
	int totalFrames;

	if (hasVideo) {
		frameRate = mpeg2VideoFrameRate(file, vstream);
		frameDelay = (int) (1000 / frameRate);
		Uint32 initialLoopStartTick, initialLoopEndTick;
		initialLoopStartTick = SDL_GetTicks();
		playReadVideoFrame(file, vstream);
		initialLoopEndTick = SDL_GetTicks();
		framesDrop =
		    (initialLoopEndTick -
		     initialLoopStartTick) / frameDelay;
		delay =
		    frameDelay - (initialLoopEndTick -
				  initialLoopStartTick) % frameDelay;
		oldDelay = initialLoopEndTick - initialLoopStartTick;
		totalFrames = mpeg2VideoGetNumFrames(file, vstream);
		printf("totalframes:%d\n", totalFrames);
	} else
		delay = 100;

	if (hasAudio)
		SDL_PauseAudio(0);

	SDL_Rect rect = { 0, 0, width, height };

	startTick = SDL_GetTicks();
	while (!end_play) {
		if (hasVideo) {
			Uint32 loopStartTick, loopEndTick;
			loopStartTick = SDL_GetTicks();

			SDL_DisplayYUVOverlay(overlay, &rect);
			playReadVideoFrame(file, vstream);

			loopEndTick = SDL_GetTicks();
			framesDrop =
			    (loopEndTick - loopStartTick) / frameDelay;
			delay =
			    frameDelay - (loopEndTick -
					  loopStartTick) % frameDelay;
			oldDelay = loopEndTick - loopStartTick;
			mpeg3_drop_frames(file, framesDrop, vstream);
			dropFrames += framesDrop;
			numFrames++;
			if (totalFrames
			    && (dropFrames + numFrames) == totalFrames)
				end_play = 1;
		}
		SDL_Delay(delay);
	}

	if (hasAudio) {
		thread_exit = 1;
		SDL_WaitThread(thread, NULL);
	}
	Uint32 endTick = SDL_GetTicks();
	printf("\n\nTotal time: %f\n",
	       (float) (endTick - startTick) / 1000.0);
	if (hasAudio) {
		printf("Total seconds read:%f\n",
		       (float) numSamples / mpeg2AudioSampleRate(file,
								 astream));
		printf("Total seconds played:%f\n",
		       (float) numSamplesPlayed /
		       mpeg2AudioSampleRate(file, astream));
		SDL_CloseAudio();
		delete stAudioCallback;
	}
	if (hasVideo) {
		printf("Total frames played:%d\n", numFrames);
		printf("Total frames droped:%d\n", dropFrames);
		SDL_FreeYUVOverlay(overlay);
	}

	SDL_Quit();
}
